/** 
 *   Site testing for SKA candidate.
 *
 *   (c) Copyright 2004 Federico Bareilles <fede@iar.unlp.edu.ar>,
 *   Instituto Argentino de Radio Astronomia (IAR).
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *     
 *   The author does NOT admit liability nor provide warranty for any
 *   of this software. This material is provided "AS-IS" in the hope
 *   that it may be useful for others.
 *
 **/
#define _GNU_SOURCE

#include <ctype.h>
#include <stdio.h>
#include <termio.h>
#include <unistd.h>
#include <sys/file.h>
#include <string.h>

#include "my_stdio.h"
#include "rotador.h"
#include "observ_table.h"


#define STR_ID      0xff

#define SET_AZ      0x80
#define SET_POL     0x81
/* #define SET_AZ_POL  0x82 */ /* not implemented in azimuth rotator */
#define SET_VEL     0x83
#define SET_CONTROL 0x84

#define GET_AZ      0x85
#define GET_POL     0x86
#define GET_STATUS  0x87
#define GET_RESET   0x88

#define POL_HORIZONTAL  1
#define POL_VERTICAL    0

/* for GET_RESET: */
#define RS_MASK   0x0f
#define RS_NEAR   0x01
#define RS_DC     0x02
#define RS_EERRO  0x04
#define RS_REMOTO 0x08

#define TO_ROTADOR_L(v) (v & 0x7f)
#define TO_ROTADOR_H(v) ((v >> 7) & 0x7f)

#define FROM_ROTADOR(h,l) (((h << 7) | l) & 0x3f)

int rotador_setup(char *name)
{
	int dev;
	struct termio term;
	
	/* Open Serial Port : 8 bits data, 1 bit stop, No parity, 2400
           baud */
	
	if ((dev = open(name, O_RDWR)) < 0 ){
		PVERB(PERR, "Cannot open serial port.\n");
		return -1;
	}
	    
	/* Zero out the baud rate portion of c_cflag */
	term.c_cflag &= ~CBAUD;
	
	/* Set baud rate to 1200 */
	term.c_cflag |= B1200;
	term.c_cflag &= ~(CSIZE|PARENB);
	term.c_cflag |= CS8 ;
  
	ioctl(dev, TCSETAW, &term);
	ioctl(dev, TCSETAW, &term);
  
	/* Turn of Canonical Processing and echo */
	term.c_lflag &= ~ICANON;
	term.c_lflag &= ~ECHO;
  
	/* Set MIN to one and TIME to zero */
	term.c_cc[VMIN]  = 1;
	term.c_cc[VTIME] = 0;
  
	/* Disable NL to CR mapping */
	term.c_iflag &= ~ICRNL;
  
	/* Set new terminal state */
	ioctl(dev, TCSETA, &term);
	ioctl(dev, TCSETA, &term);

	rotador_set_control(dev, CT_AZ_EN | CT_PO_EN | CT_PW_ON);
	
	return dev;
}


ssize_t soft_read_time_out(struct timeval tv, int fd, void *buf, size_t count)
{
	fd_set rfds;
	int fd_max = fd + 1;
	int retval;

	FD_ZERO(&rfds);
	FD_SET(fd, &rfds);
	retval = select(fd_max, &rfds, NULL, NULL, &tv);
	if (retval) {
		if( FD_ISSET(fd, &rfds) ) {
			return read(fd, buf, count);
		}
	}

	return -1;
}


int rotador_wait_confitmation(int fd)
{
	struct timeval tv;
	char answer[5];

	tv.tv_sec = ROTATOR_TIME_OUT;
	tv.tv_usec = 0;	
	if( soft_read_time_out(tv, fd, answer, 5) < 0 ) {
		SYSLOG(LOG_WARNING, "azimuth rotator time out.");
	}else{
			/* TODO: No esta definido aun por Esteban */
		return 0;
	}

	return -1;
}


int discard_noise(int fd)
{
	struct timeval tv;
	char noise;

	tv.tv_sec = 0;
	tv.tv_usec = 0;
	while( soft_read_time_out(tv, fd, &noise, 1) > 0); 

	return 0;
}


ssize_t soft_read(int fd, void *buf, size_t count)
{
	struct timeval tv;

	tv.tv_sec = 5;
	tv.tv_usec = 0;

	return soft_read_time_out(tv, fd, buf, count);
}


ssize_t soft_write(int fd, const void *buf, size_t count)
{
	discard_noise(fd);
	return write(fd, buf, count);
}


int rotador_set_azimut(int fd, int az)
{
	int const len = 5;
	char answer[len];
	char cmd[len];
	
	cmd[0] = STR_ID;
	cmd[1] = SET_AZ;
	cmd[2] = TO_ROTADOR_H(az);
	cmd[3] = TO_ROTADOR_L(az);
	cmd[4] = cmd[1] | cmd[2] | cmd[3];
	soft_write(fd, cmd, len);
	soft_read(fd, answer, len);

	if( strncmp( answer, cmd, len) != 0) {
		return -1;
	}

	return rotador_wait_confitmation(fd);
}


int rotador_set_polarization(int fd, char pol)
{
	int const len = 4;
	char answer[len];
	char cmd[len];
	char poln;

	if (pol == HORIZONTAL) poln = POL_HORIZONTAL;
	else if (pol == VERTICAL) poln = POL_VERTICAL;
	else return -1;
	
	cmd[0] = STR_ID;
	cmd[1] = SET_POL;
	cmd[2] = poln;
	cmd[3] = cmd[1] | cmd[2];
	soft_write(fd, cmd, len);
	soft_read(fd, answer, len);

	if( strncmp( answer, cmd, len) != 0) {
		return -1;
	}

	return rotador_wait_confitmation(fd);
}


int rotador_set_velocity(int fd, int vel)
{
	int const len = 4;
	char answer[len];
	char cmd[len];
	
	cmd[0] = STR_ID;
	cmd[1] = SET_VEL;
	cmd[2] = TO_ROTADOR_L(vel);
	cmd[3] = cmd[1] | cmd[2];
	soft_write(fd, cmd, len);
	soft_read(fd, answer, len);

	if( strncmp( answer, cmd, len) != 0) {
		return -1;
	}

	return 0;
}



int rotador_get_azimut(int fd, int *az)
{
	int const len = 5;
	char answer[len];
	char cmd[3];
	
	cmd[0] = STR_ID;
	cmd[1] = GET_AZ;
	cmd[2] = GET_AZ;
	soft_write(fd, cmd, 3);
	soft_read(fd, answer, len);
	
	if( (answer[1] | answer[2] | answer[3]) != answer[4] ) {
		SYSLOG(LOG_ERR, "Bad packet from rotator.");
		return -1;
	}
	*az = FROM_ROTADOR(answer[2], answer[3]);

	return 0;
}



int rotador_get_polarization(int fd, char *pol)
{
	int const len = 4;
	char answer[len];
	char cmd[3];
	
	cmd[0] = STR_ID;
	cmd[1] = GET_POL;
	cmd[2] = GET_POL;
	soft_write(fd, cmd, 3);
	soft_read(fd, answer, len);
	
	if( (answer[1] | answer[2]) != answer[3] ) {
		SYSLOG(LOG_ERR, "Bad packet from rotator.");
		return -1;
	}
	if(answer[2] == POL_HORIZONTAL)
		*pol = HORIZONTAL;
	else
		*pol = VERTICAL;

	return 0;
}


int rotador_set_control(int fd, char ctrl)
{
	int const len = 4;
	char answer[len];
	char cmd[len];
	
	cmd[0] = STR_ID;
	cmd[1] = SET_CONTROL;
	cmd[2] = ctrl & CT_MASK;
	cmd[3] = cmd[1] | cmd[2];
	soft_write(fd, cmd, len);
	soft_read(fd, answer, len);

	if( strncmp( answer, cmd, len) != 0) {
		return -1;
	}

	return 0;
}


int rotador_get_status(int fd, char *status)
{
	int const len = 4;
	char answer[len];
	char cmd[3];
	
	cmd[0] = STR_ID;
	cmd[1] = GET_STATUS;
	cmd[2] = GET_STATUS;
	soft_write(fd, cmd, 3);
	soft_read(fd, answer, len);
	
	if( (answer[1] | answer[2]) != answer[3] ) {
		SYSLOG(LOG_ERR, "Bad packet from rotator.");
		return -1;
	}
	*status = answer[2];

	return 0;
}
